@@ -1,7 +1,6 @@
-from .base import TrainingCallback, PredictionCallback, CallbackException
+from .base import TrainingCallback, PredictionCallback, CallbackException, ValueTrainingCallback
 from .early_stopping import EarlyStopping
 from .logging import EpochProgressLogger, BatchProgressLogger, PredictionProgressLogger
-from .parallel import SynchronizationCallback, ModelSharingCallback
 from .saving import ModelSaverCallback
 from .schedule import LearningRateScheduler, ParameterScheduler
 from .tracking import NeptuneTracker, TensorboardTracker
