@@ -1,17 +1,17 @@
-import os
-import time
 from abc import ABC, abstractmethod
+from collections import defaultdict
 import torch
 import torch.nn as nn
 import torch.cuda as cuda
 import pyblaze.multiprocessing as xmp
-from pyblaze.nn.callbacks import TrainingCallback, PredictionCallback, CallbackException
+from pyblaze.nn.callbacks import TrainingCallback, PredictionCallback, CallbackException, \
+    ValueTrainingCallback
 import pyblaze.nn.utils as xnnu
 from pyblaze.utils.torch import gpu_device, to_device
-from pyblaze.utils.stdlib import flatten
-from .wrappers import History, Evaluation
+from ._history import History
+from ._utils import forward
 
-class BaseEngine(TrainingCallback, PredictionCallback, ABC):
+class Engine(TrainingCallback, PredictionCallback, ABC):
     """
     A base class for training and evaluating models as well as making predictions. Generally, this
     class should be seen as _binding_ between a model and some data. Models should always be
@@ -22,9 +22,9 @@ class BaseEngine(TrainingCallback, PredictionCallback, ABC):
     independent, identically distributed data samples) and/or model types (e.g. GANs).
     """
 
-    #########################
-    ### ENGINE PROPERTIES ###
-    #########################
+    ################################################################################
+    ### ENGINE CONFIGURATION
+    ################################################################################
     def supports_multiple_gpus(self):
         """
         Returns whether the engine allows multiple GPUs to be used during training. By default, it
@@ -37,9 +37,9 @@ class BaseEngine(TrainingCallback, PredictionCallback, ABC):
         """
         return True
 
-    ######################
-    ### INITIALIZATION ###
-    ######################
+    ################################################################################
+    ### INITIALIZATION
+    ################################################################################
     def __init__(self, model):
         """
         Initializes a new engine for a specified model.
@@ -54,22 +54,21 @@ class BaseEngine(TrainingCallback, PredictionCallback, ABC):
         self._cache = {}
         self._iteration = None
 
-    ##########################
-    ### HIGH-LEVEL METHODS ###
-    ##########################
+    ################################################################################
+    ### TRAINING
+    ################################################################################
     # pylint: disable=too-many-branches,too-many-statements
-    def train(self, train_data, val_data=None, epochs=20, val_iterations=None,
-              eval_every=None, eval_train=False, eval_val=True, callbacks=None,
-              metrics=None, gpu='auto', **kwargs):
-        """
+    def train(self, train_data, val_data=None, epochs=20, val_iterations=None, eval_every=None,
+              eval_train=False, eval_val=True, callbacks=None, metrics=None, gpu='auto', **kwargs):
+        r"""
         Method for training the model with the supplied parameters.
 
         Parameters
         ----------
-        train_data: torch.DataLoader
+        train_data: torch.utils.data.DataLoader
             A data loader to obtain training data from. The samples yielded by the data loader
-            depend on a specific trainer implementation.
-        val_data: torch.DataLoader, default: None
+            depend on a specific engine implementation.
+        val_data: torch.utisl.data..DataLoader, default: None
             A data loader to use for validation. If the loader is an infinite data loader,
             `val_iterations` must also be given. If not supplied, no validation will be performed.
         epochs: int, default: 20
@@ -79,19 +78,19 @@ class BaseEngine(TrainingCallback, PredictionCallback, ABC):
             The number of iterations to perform for validation. Must be given only if `val_data` is
             an infinite data loader. Otherwise, the parameter is ignored.
         eval_every: int, default: None
-            How many iterations should pass until validation is called again. This should only be
-            set if `train_data` is based on an iterable dataset which has infinite length. Note
-            that everything works as if `epochs // eval_every` epochs is trained while eval_every
-            can be considered the number of mini-batches.
+            How many iterations should pass until validation is called again. If this is set when
+            `train_data` is an iterable dataset, then it defines the number of iterations per
+            epoch (i.e. before performing validation again). Otherwise, it defines the number of
+            epochs to train before calling validation.
         eval_train: bool, default: False
             Whether to compute metrics (apart from the loss) for both the validation and the train
             data. If this flag is set to False, metrics will only be computed for the validation
             data.
         eval_val: bool, default: True
             Whether to compute the loss (apart from the metrics) for the validation data. If this
-            flag is set to False, no validation loss will be computed.
-        callbacks: list of pyblaze.nn.TrainingCallback or
-                pyblaze.nn.PredictionCallback, default: []
+            flag is set to `False`, no validation loss will be computed.
+        callbacks: list of (pyblaze.nn.TrainingCallback or pyblaze.nn.PredictionCallback),
+                default: []
             The callbacks to use for training and inference. The training callbacks and prediction
             callbacks will be filtered automatically.
         metrics: dict of str -> func, default: {}
@@ -106,16 +105,18 @@ class BaseEngine(TrainingCallback, PredictionCallback, ABC):
             accordingly. If set to a string, the only valid value is 'auto'. In this case, all
             available GPUs are used.
         kwargs: keyword arguments
-            Additional keyword arguments dependent on the specific subclass.
+            Additional keyword arguments dependent on the specific subclass. If prefixed with
+            'eval\_', it will be passed to :meth:`eval_batch` (without the prefix), otherwise to
+            :meth:`train_batch`. All keyword arguments may also be given as `ParameterScheduler`
+            to enable dynamic changes in parameters. Note that the scheduler can also be used to
+            alter values for 'eval\_' keywords. However, they do so following the same schedule as
+            the training keywords. In any case, consult the engine's documentation if the
+            parameter allows dynamic behavior. Semantically, it might not make sense to do.
 
         Returns
         -------
         pyblaze.nn.History
-            A history object summarizing stats from the training. It contains as properties the
-            development of the loss as `train_loss` (and potentially `val_loss`, if `val_data` is
-            supplied and the keyword arguments include a function called `loss`). If additional
-            metrics are supplied, there will be a property `val_<metric>` for each metric, and
-            potentially `train_<metric>` if `eval_train` is set to True.
+            A history object summarizing stats from the training.
         """
         if metrics is None:
             metrics = {}
@@ -123,34 +124,37 @@ class BaseEngine(TrainingCallback, PredictionCallback, ABC):
             callbacks = []
 
         # 1) Setup
-        exception = None
-        tic = time.time()
+        try:
+            batch_iterations = len(train_data)
+            iterable_data = False
+        except: # pylint: disable=bare-except
+            batch_iterations = eval_every
+            iterable_data = True
 
-        if eval_every is not None:
+        exception = None
+        if iterable_data and eval_every is not None:
             # Here, epochs are considered iterations
             epochs = epochs // eval_every
 
         # 1.1) Callbacks
-        train_callbacks = [
-            c for c in callbacks if isinstance(c, TrainingCallback)
-        ]
-        prediction_callbacks = [
-            c for c in callbacks if isinstance(c, PredictionCallback)
-        ]
-
-        self._exec_callbacks(
-            train_callbacks, 'before_training', self.model, epochs
-        )
+        history = History()
+        # Prepend the engine's callbacks to the passed callbacks
+        callbacks = [history] + callbacks
+        # Also, add the callbacks that are extracted from the keyword arguments
+        callbacks += [v for _, v in kwargs.items() if isinstance(v, TrainingCallback)]
+        # Then, we can extract the callbacks for training and prediction
+        train_callbacks = [c for c in callbacks if isinstance(c, TrainingCallback)]
+        prediction_callbacks = [c for c in callbacks if isinstance(c, PredictionCallback)]
+        self._exec_callbacks(train_callbacks, 'before_training', self.model, epochs)
 
         # 1.2) Metrics
-        metric_history = []
         if eval_val and 'loss' in kwargs:
             val_metrics = {**metrics, **{'loss': kwargs['loss']}}
         else:
             val_metrics = metrics
 
         # 1.3) Data loading
-        if eval_every is not None:
+        if iterable_data:
             train_iterator = iter(train_data)
 
         # 1.4) GPU support
@@ -158,15 +162,22 @@ class BaseEngine(TrainingCallback, PredictionCallback, ABC):
         self._setup_device(gpu)
         self.model.to(self.device)
 
+        # 1.5) Valid kwargs
+        train_kwargs = {k: v for k, v in kwargs.items() if not k.startswith('eval_')}
+        dynamic_train_kwargs = {
+            k: v for k, v in train_kwargs.items() if isinstance(v, ValueTrainingCallback)
+        }
+        eval_kwargs = {k[5:]: v for k, v in kwargs.items() if k.startswith('eval_')}
+        dynamic_eval_kwargs = {
+            k: v for k, v in eval_kwargs.items() if isinstance(v, ValueTrainingCallback)
+        }
+
         # 2) Train for number of epochs
         for current_epoch in range(epochs):
             # 2.1) Prepare
-            batch_iterations = eval_every or len(train_data)
-
             try:
                 self._exec_callbacks(
-                    train_callbacks, 'before_epoch', current_epoch,
-                    batch_iterations
+                    train_callbacks, 'before_epoch', current_epoch, batch_iterations
                 )
             except CallbackException as e:
                 exception = e
@@ -175,61 +186,51 @@ class BaseEngine(TrainingCallback, PredictionCallback, ABC):
             # 2.2) Train
             self.model.train()
 
-            train_batch_weights = []
-            train_losses = []
+            batch_losses = []
+            if not iterable_data:
+                train_iterator = iter(train_data)
 
-            if eval_every is not None:
-                # 2.2.1) Iterable dataset
-                for _ in range(eval_every):
-                    item = next(train_iterator)
-                    item = to_device(self.device, item)
-                    loss = self.train_batch(item, **kwargs)
-                    train_losses.append(loss)
-                    self._exec_callbacks(
-                        train_callbacks, 'after_batch', loss
-                    )
-            else:
-                # 2.2.2) Dataset
-                for item in train_data:
-                    item = to_device(self.device, item)
-                    loss = self.train_batch(item, **kwargs)
-                    train_batch_weights.append(len(item))
-                    train_losses.append(loss)
-                    self._exec_callbacks(
-                        train_callbacks, 'after_batch', loss
-                    )
+            for _ in range(batch_iterations):
+                train_kwargs = {
+                    **train_kwargs,
+                    **{k: v.read() for k, v in dynamic_train_kwargs.items()}
+                }
+                item = next(train_iterator)
+                item = to_device(self.device, item)
+                loss = self.train_batch(item, **train_kwargs)
+                batch_losses.append(loss)
+                self._exec_callbacks(train_callbacks, 'after_batch', _strip_metrics(loss))
 
             # 2.3) Validate
-            batch_metrics = Evaluation(self.collate_losses(train_losses), train_batch_weights)
+            epoch_metrics = self.collate_losses(batch_losses)
+            eval_kwargs = {
+                **eval_kwargs,
+                **{k: v.read() for k, v in dynamic_eval_kwargs.items()}
+            }
+            do_val = eval_every is None or iterable_data or \
+                current_epoch % eval_every == 0 or current_epoch == epochs - 1
 
-            if val_data is not None:
-                eval_val = self.evaluate(
+            if val_data is not None and do_val:
+                eval_metrics = self.evaluate(
                     val_data, iterations=val_iterations, metrics=val_metrics,
-                    callbacks=prediction_callbacks, gpu=None
-                ).with_prefix('val_')
-                batch_metrics = Evaluation.merge(batch_metrics, eval_val)
+                    callbacks=prediction_callbacks, gpu=None, **eval_kwargs
+                )
+                epoch_metrics = {
+                    **epoch_metrics, **{f'val_{k}': v for k, v in eval_metrics.items()}
+                }
 
-            if eval_train:
-                eval_train = self.evaluate(
+            if eval_train and do_val:
+                eval_metrics = self.evaluate(
                     train_data, iterations=val_iterations, metrics=metrics,
-                    callbacks=prediction_callbacks, gpu=None
-                ).with_prefix('train_')
-                batch_metrics = Evaluation.merge(batch_metrics, eval_train)
-
-            batch_metrics = Evaluation.merge(
-                batch_metrics, Evaluation({'_timestamp': time.time()})
-            )
-
-            metric_history.append({
-                'micro_train_loss': train_losses,
-                **batch_metrics.to_dict()
-            })
+                    callbacks=prediction_callbacks, gpu=None, **eval_kwargs
+                )
+                epoch_metrics = {
+                    **epoch_metrics, **{f'train_{k}': v for k, v in eval_metrics.items()}
+                }
 
             # 2.4) Finish epoch
             try:
-                self._exec_callbacks(
-                    train_callbacks, 'after_epoch', batch_metrics
-                )
+                self._exec_callbacks(train_callbacks, 'after_epoch', epoch_metrics)
             except CallbackException as e:
                 exception = e
                 break
@@ -241,18 +242,19 @@ class BaseEngine(TrainingCallback, PredictionCallback, ABC):
             self.device = None
 
         # 3.2) Finish callbacks
-        self._exec_callbacks(
-            train_callbacks, 'after_training'
-        )
+        self._exec_callbacks(train_callbacks, 'after_training')
         if exception is not None:
             if isinstance(exception, CallbackException):
                 exception.print()
             else:
                 print(exception)
 
-        return History(time.time() - tic, metric_history)
+        return history
 
-    def evaluate(self, data, iterations=None, metrics=None, callbacks=None, gpu='auto'):
+    ################################################################################
+    ### EVALUATION
+    ################################################################################
+    def evaluate(self, data, iterations=None, metrics=None, callbacks=None, gpu='auto', **kwargs):
         """
         Evaluates the model on the given data and computes the supplied metrics.
 
@@ -260,7 +262,7 @@ class BaseEngine(TrainingCallback, PredictionCallback, ABC):
         ----------
         data: torch.DataLoader
             A data loader to obtain evaluation samples from. The expected samples depend on a
-            specific trainer subclass.
+            specific engine subclass.
         iterations: int, default: None
             The number of samples used for evaluating if the given data is an infinite data loader.
         metrics: dict of str -> func, default: {}
@@ -276,6 +278,8 @@ class BaseEngine(TrainingCallback, PredictionCallback, ABC):
             simultaneously. In this case, the batch sizes of the data loaders should be adjusted
             accordingly. If set to 'auto', all available GPUs are used. In case of `None`, the
             model will not be moved. Only use this option if you know what you are doing.
+        kwargs: keyword arguments
+            Additional arguments, passed directly to the `eval_batch` method.
 
         Returns
         -------
@@ -287,63 +291,48 @@ class BaseEngine(TrainingCallback, PredictionCallback, ABC):
         if callbacks is None:
             callbacks = []
 
+        # Setup
         num_predictions = iterations or len(data)
-        self._exec_callbacks(
-            callbacks, 'before_predictions', self.model, num_predictions
-        )
+        self._exec_callbacks(callbacks, 'before_predictions', self.model, num_predictions)
 
+        # Ensure GPU
         if gpu is not None:
             gpu = self._gpu_descriptor(gpu)
             self._setup_device(gpu)
             self.model.to(self.device)
 
+        # Run inference
         self.model.eval()
 
-        predictions = []
-        targets = []
-
+        evals = []
         iterator = iter(data)
         for _ in range(num_predictions):
             item = next(iterator)
             item = to_device(self.device, item)
 
             with torch.no_grad():
-                prediction, target = self.eval_batch(item)
+                eval_out = self.eval_batch(item, **kwargs)
 
-            predictions.append(to_device('cpu', prediction))
-            targets.append(to_device('cpu', target))
+            evals.append(to_device('cpu', eval_out))
+            self._exec_callbacks(callbacks, 'after_batch', None)
 
-            self._exec_callbacks(
-                callbacks, 'after_batch', None
-            )
+        self._exec_callbacks(callbacks, 'after_predictions')
 
-        self._exec_callbacks(
-            callbacks, 'after_predictions'
-        )
-
-        predictions = self.collate_predictions(predictions)
-        targets = self.collate_targets(targets)
-
-        def process_metric(k, m):
-            if isinstance(m, dict):
-                return [(f'{k}_{mk}', mm.item()) for mk, mm in m.items()]
-            return [(k, m.item())]
-
-        if targets is None:
-            result = predictions
-        else:
-            result = dict(flatten(
-                process_metric(k, f(predictions, targets))
-                for k, f in metrics.items()
-            ))
+        evals = self.collate_evals(evals)
 
         if gpu is not None:
             self.model.to('cpu', non_blocking=True)
             self.device = None
 
-        return Evaluation(result)
+        return {
+            key: self._process_metric(forward(metric, evals))
+            for key, metric in metrics.items()
+        }
 
-    def predict(self, data, iterations=None, callbacks=None, gpu='auto', **kwargs):
+    ################################################################################
+    ### PREDICTIONS
+    ################################################################################
+    def predict(self, data, iterations=None, callbacks=None, gpu='auto', parallel=True, **kwargs):
         """
         Computes predictions for the given samples.
 
@@ -364,6 +353,10 @@ class BaseEngine(TrainingCallback, PredictionCallback, ABC):
             multiple GPUs makes up for this overhead. If `False` is specified, all cores of the
             computer are used to make predictions in parallel. In the case of 'auto', all available
             GPUs are used (if any).
+        parallel: bool, default: True
+            Whether to allow predictions to be done in parallel. Normally, you do not need to set
+            this to `False`, unless there is some good reason (e.g. running in a Docker container
+            that does not provide sufficient shared memory).
         kwargs: keyword arguments
             Additional arguments passed directly to the `predict_batch` function.
 
@@ -381,9 +374,7 @@ class BaseEngine(TrainingCallback, PredictionCallback, ABC):
         # 2) Setup data loading
         num_iterations = iterations or len(data)
 
-        self._exec_callbacks(
-            callbacks, 'before_predictions', self.model, num_iterations
-        )
+        self._exec_callbacks(callbacks, 'before_predictions', self.model, num_iterations)
 
         # 3) Make sure that the model is not data parallel, we don't need this for predicting
         if isinstance(self.model, nn.DataParallel):
@@ -394,26 +385,26 @@ class BaseEngine(TrainingCallback, PredictionCallback, ABC):
         # 4) Now perform predictions
         if (isinstance(gpu, list) and len(gpu) > 1) or not gpu:
             # parallel computation
-            if isinstance(gpu, list):
+            if parallel and isinstance(gpu, list):
                 num_workers = len(gpu)
-            elif isinstance(gpu, bool):
-                num_workers = os.cpu_count()
+            elif parallel and isinstance(gpu, bool):
+                # number of CPU cores
+                num_workers = -1
             else:
-                num_workers = 1
+                # execute on main thread otherwise
+                num_workers = 0
 
             model.share_memory()
 
             callback = lambda: self._exec_callbacks(callbacks, 'after_batch', None)
             vectorizer = xmp.Vectorizer(
-                _prediction_worker_func, _prediction_worker_init,
-                callback_func=callback, num_workers=num_workers,
-                gpu=gpu, model=model
+                _prediction_worker_func, _prediction_worker_init, callback_func=callback,
+                num_workers=num_workers, gpu=gpu, model=model
             )
 
             iterator = iter(data)
             predictions = vectorizer.process(
-                (next(iterator) for _ in range(num_iterations)),
-                self.predict_batch, kwargs
+                (next(iterator) for _ in range(num_iterations)), self.predict_batch, kwargs
             )
         else:
             # sequential computation
@@ -426,23 +417,17 @@ class BaseEngine(TrainingCallback, PredictionCallback, ABC):
             for _ in range(num_iterations):
                 x = next(iterator)
 
-                out = _prediction_worker_func(
-                    x, self.predict_batch, kwargs, device
-                )
+                out = _prediction_worker_func(x, self.predict_batch, kwargs, device)
                 predictions.append(out)
-                self._exec_callbacks(
-                    callbacks, 'after_batch', None
-                )
+                self._exec_callbacks(callbacks, 'after_batch', None)
 
-        self._exec_callbacks(
-            callbacks, 'after_predictions'
-        )
+        self._exec_callbacks(callbacks, 'after_predictions')
 
         return self.collate_predictions(predictions)
 
-    ########################
-    ### BATCH PROCESSING ###
-    ########################
+    ################################################################################
+    ### BATCH PROCESSING
+    ################################################################################
     @abstractmethod
     def train_batch(self, data, **kwargs):
         """
@@ -458,7 +443,9 @@ class BaseEngine(TrainingCallback, PredictionCallback, ABC):
         -------
         object
             The loss computed for the batch. If the returned value is not float, overwrite
-            `collate_losses`.
+            :meth:`collate_losses`. If you return a dict here, all keys prefixed with an underscore
+            will not be passed to any callbacks but only be available in the :meth:`collate_losses`
+            method.
         """
 
     @abstractmethod
@@ -477,11 +464,9 @@ class BaseEngine(TrainingCallback, PredictionCallback, ABC):
 
         Returns
         -------
-        torch.Tensor
-            The output from the model.
         object
-            The target, i.e. correct output. If this is not a `torch.Tensor`, the `collate_targets`
-            should be overriden.
+            The model output and possibly some correct target. If this is not a
+            :code:`torch.Tensor`, the :meth:`collate_evals`.
         """
 
     # pylint: disable=unused-argument
@@ -506,50 +491,22 @@ class BaseEngine(TrainingCallback, PredictionCallback, ABC):
         object
             The return value passed to `collate_predictions`.
         """
-        return self.forward(data)
-
-    ###########################
-    ### COLLATION FUNCTIONS ###
-    ###########################
-    def collate_predictions(self, predictions):
-        """
-        Combines the predictions obtained from the `eval_batch` function. The default
-        implementation assumes that predictions are tensors and can simply be conatenated.
-
-        Parameters
-        ----------
-        predictions: list of objects
-            The predictions.
-
-        Returns
-        -------
-        object
-            An object to be used as predicted value for some metric.
-        """
-        return torch.cat(predictions)
-
-    def collate_targets(self, targets):
-        """
-        Combines the targets in a way that it can be passed to some metric. The default
-        implementation assumes that targets are tensors and simply concatenates them.
-
-        Parameters
-        ----------
-        list of object
-            The targets.
-
-        Returns
-        -------
-        object
-            An object to be used as true target with some metric.
-        """
-        return torch.cat(targets)
+        return forward(self.model, data)
 
+    ################################################################################
+    ### COLLATION FUNCTIONS
+    ################################################################################
     def collate_losses(self, losses):
         """
-        Combines the losses obtained from the `train_batch` function. The default implementation
+        Combines the losses obtained from the :meth:`train_batch` method. The default implementation
         assumes that simple floats are returned. Overwriting this method might be useful if you
-        have different types of losses, such as generator and discriminator loss when training GANs.
+        have different types of losses, such as generator and critic loss when training Wasserstein
+        GANs.
+
+        Note
+        ----
+        Overwrite this function if it is important to weigh losses obtained from batches in some
+        manner (e.g. if data batches have different size).
 
         Parameters
         ----------
@@ -558,34 +515,67 @@ class BaseEngine(TrainingCallback, PredictionCallback, ABC):
 
         Returns
         -------
-        dict of str -> (float or list of float)
-            The loss names mapped to their values.
+        dict of str -> object
+            The loss names mapped to their values (usually float values).
         """
-        return {'train_loss': losses}
+        ref = losses[0]
 
-    ######################
-    ### BASE FUNCTIONS ###
-    ######################
-    def forward(self, x):
+        if isinstance(ref, dict):
+            result = defaultdict(list)
+            for item in losses:
+                for k, v in item.items():
+                    result[k].append(v)
+            return {k: sum(v) / len(v) for k, v in result.items()}
+
+        if isinstance(ref, (list, tuple)):
+            result = [[] for _ in len(ref)]
+            for item in losses:
+                for i, it in enumerate(item):
+                    result[i].append(it)
+            return {f'loss_{i}': sum(r) / len(r) for i, r in enumerate(result)}
+
+        return {'loss': sum(losses) / len(losses)}
+
+    def collate_evals(self, evals):
         """
-        Computes the model output for common types of inputs.
+        Combines the evaluation objects returned from the :meth:`eval_batch` method. The default
+        implementation works whenever the returned objects are tensors, tuples of tensors or
+        dictionaries of tensors. The contents will then simply be concatenated.
 
         Parameters
         ----------
-        x: list or tuple or dict or object
-            The input the model.
+        evals: list of objects
+            The evaluation objects.
 
         Returns
         -------
         object
-            The output of the model (although arbitrary, it is usually a torch.Tensor).
+            An object that is passed to all metrics. If a tuple or a dict, they are passed as
+            variadic arguments or keyword arguments, respectively.
         """
-        if isinstance(x, (list, tuple)):
-            return self.model(*x)
-        if isinstance(x, dict):
-            return self.model(**x)
-        return self.model(x)
+        return self._collate(evals)
 
+    def collate_predictions(self, predictions):
+        """
+        Combines the predictions obtained for multiple batches as per the :meth:`predict_batch`
+        method. The default implementation works whenever the returned objects are tensors, tuples
+        of tensors or dictionaries of tensors. The contents will then simply be concatenated.
+
+        Parameters
+        ----------
+        predictions: list of objects
+            The predictions obtained.
+
+        Returns
+        -------
+        object
+            An object that is obtained from the predictions.
+        """
+        return self._collate(predictions)
+
+    ################################################################################
+    ### UTILITY FUNCTIONS
+    ################################################################################
     def _gpu_descriptor(self, gpu):
         if gpu == 'auto':
             if cuda.device_count() == 0:
@@ -606,6 +596,29 @@ class BaseEngine(TrainingCallback, PredictionCallback, ABC):
         for callback in [self] + callbacks:
             getattr(callback, func)(*args)
 
+    def _collate(self, items):
+        ref = items[0]
+        if isinstance(ref, dict):
+            return {key: torch.cat([v[key] for v in items]) for key in ref.keys()}
+        # recursive call of _collate for a more generic functionality
+        if isinstance(ref, (list, tuple)):
+            return tuple(self._collate([v[i] for v in items]) for i in range(len(ref)))
+        return torch.cat(items)
+
+    def _process_metric(self, metric):
+        if isinstance(metric, torch.Tensor):
+            if metric.numel() == 1:
+                return metric.item()
+            return metric
+        return metric
+
+
+def _strip_metrics(metrics):
+    # If `metrics` is a dict, we remove all keys starting with an underscore
+    if isinstance(metrics, dict):
+        return {k: v for k, v in metrics.items() if not k.startswith('_')}
+    return metrics
+
 
 def _prediction_worker_func(item, predict, kwargs, device):
     x = to_device(device, item)
